// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test4mat.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	06.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 4:
// p3(X, Y) :- q(X, Y), p2(Y, Z).
// p2(X, Y) :- q(X, Y), p1(Y, Z).
// p1(X, Y) :- q(X, Y), p0(Y, Z).
// p0(X, Y) :- r(X, Y).
//
// q := {(i,i+1)|i = 1,...,50000}
// r := {(i,i)  |i = 1,...,50000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 1000000;

const int R_TUPLES = 1000000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Predicate p0 - implemented as linked list
// ----------------------------------------------------------------------------

class p0 {
	friend class p0_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		p0 *new_elem = new p0(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p0(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p0 *Next;

	static class p0* First;
	static class p0* Last;
};

class p0* p0::First;
class p0* p0::Last;

class p0_cursor {
public:
	p0_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p0::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p0*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p1 - implemented as linked list
// ----------------------------------------------------------------------------

class p1 {
	friend class p1_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		p1 *new_elem = new p1(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p1(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p1 *Next;

	static class p1* First;
	static class p1* Last;
};

class p1* p1::First;
class p1* p1::Last;

class p1_cursor {
public:
	p1_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p1::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p1*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p2 - implemented as linked list
// ----------------------------------------------------------------------------

class p2 {
	friend class p2_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		p2 *new_elem = new p2(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p2(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p2 *Next;

	static class p2* First;
	static class p2* Last;
};

class p2* p2::First;
class p2* p2::Last;

class p2_cursor {
public:
	p2_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p2::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p2*   Current;
};


// ----------------------------------------------------------------------------
// Predicate p3 - implemented as linked list
// ----------------------------------------------------------------------------

class p3 {
	friend class p3_cursor;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2) {
		p3 *new_elem = new p3(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

private:
	p3(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p3 *Next;

	static class p3* First;
	static class p3* Last;
};

class p3* p3::First;
class p3* p3::Last;

class p3_cursor {
public:
	p3_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p3::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() const { return Current->col_1(); }
	int col_2() const { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p3*   Current;
};

//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() const { return i; }

	int col_2() const { return i+1; }

	void close() {}

private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r: Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor {
public:
	r_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < R_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() const { return i; }

	int col_2() const { return i; }

	void close() {}

private:
	int i;
};

//-----------------------------------------------------------------------------
// Rule Implementation: Materialization with Merge Join
//-----------------------------------------------------------------------------

void exec_rule4()
{
	// p0(X, Y) :- r(X, Y).
	r_cursor r_c;
	r_c.open();
	while(r_c.fetch()) {
		int X = r_c.col_1();
		int Y = r_c.col_2();
		p0::insert(X, Y);
	}
}

void exec_rule3_alt()
{

	// p1(X, Z) :- q(X, Y), p0(Y, Z).
	q_cursor left;
	p0_cursor right;
	left.open();
	right.open();
	bool left_more = left.fetch();
	bool right_more = right.fetch();
	while(left_more && right_more) {
		int X = left.col_1();
		int Y_left = left.col_2();
		int Y_right = right.col_1();
		int Z = right.col_2();
		if(Y_left == Y_right) {
			p1::insert(X, Z);
			left_more = left.fetch();
			right_more = right.fetch();
		}
		else if(Y_left < Y_right) {
			left_more = left.fetch();
		}
		else { // Y_left > Y_right
			right_more = right.fetch();
		}
	}
}

void exec_rule3()
{

	// p1(X, Z) :- q(X, Y), p0(Y, Z).

	q_cursor left;
	p0_cursor right;
	int X, Y_left, Y_right, Z;

	left.open();
	right.open();
	if(!left.fetch())
		return;
	if(!right.fetch())
		return;
	X = left.col_1();
	Y_left = left.col_2();
	Y_right = right.col_1();
	Z = right.col_2();

	while(1) {
		if(Y_left == Y_right) {
			p1::insert(X, Z);
			if(!left.fetch())
				return;
			if(!right.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
			Y_right = right.col_1();
			Z = right.col_2();
		}
		else if(Y_left < Y_right) {
			if(!left.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
		}
		else { // Y_left > Y_right
			if(!right.fetch())
				return;
			Y_right = right.col_1();
			Z = right.col_2();
		}
	}
}

void exec_rule2()
{
	// p2(X, Z) :- q(X, Y), p1(Y, Z).

	q_cursor left;
	p1_cursor right;
	int X, Y_left, Y_right, Z;

	left.open();
	right.open();
	if(!left.fetch())
		return;
	if(!right.fetch())
		return;
	X = left.col_1();
	Y_left = left.col_2();
	Y_right = right.col_1();
	Z = right.col_2();
	while(1) {
		if(Y_left == Y_right) {
			p2::insert(X, Z);
			if(!left.fetch())
				return;
			if(!right.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
			Y_right = right.col_1();
			Z = right.col_2();
		}
		else if(Y_left < Y_right) {
			if(!left.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
		}
		else { // Y_left > Y_right
			if(!right.fetch())
				return;
			Y_right = right.col_1();
			Z = right.col_2();
		}
	}
}

void exec_rule1()
{

	// p3(X, Z) :- q(X, Y), p2(Y, Z).

	q_cursor left;
	p2_cursor right;
	int X, Y_left, Y_right, Z;

	left.open();
	right.open();
	if(!left.fetch())
		return;
	if(!right.fetch())
		return;
	X = left.col_1();
	Y_left = left.col_2();
	Y_right = right.col_1();
	Z = right.col_2();
	while(1) {
		if(Y_left == Y_right) {
			p3::insert(X, Z);
			if(!left.fetch())
				return;
			if(!right.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
			Y_right = right.col_1();
			Z = right.col_2();
		}
		else if(Y_left < Y_right) {
			if(!left.fetch())
				return;
			X = left.col_1();
			Y_left = left.col_2();
		}
		else { // Y_left > Y_right
			if(!right.fetch())
				return;
			Y_right = right.col_1();
			Z = right.col_2();
		}
	}
}

void exec_prog()
{
	// Compute p0:
	exec_rule4();

	// Compute p1:
	exec_rule3();

	// Compute p2:
	exec_rule2();

	// Compute p1:
	exec_rule1();
}

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;

	cout << "Test of Method 1: Materialization+Merge Join [Program 4]\n";
	t.start();
	exec_prog();
	p3_cursor p3_c;
	p3_c.open();
	while(p3_c.fetch()) {
		c1 += p3_c.col_1();
		c2 += p3_c.col_2();
	}
	p3_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		int num_tuples = 0;
		cout << "Extension of p3:\n\n";
		p3_c.open();
		while(p3_c.fetch()) {
			num_tuples++;
			cout << "p3(" << p3_c.col_1() << ", " << p3_c.col_2()
				<< ").\n";
		}
		cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
		p3_c.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


